"""Módulo que define el cliente para comunicación con BigQuery."""

# Librerías Externas.
from typing import Optional, Dict, List

import logging

import pandas as pd

from google.cloud import bigquery

# Librerías Internas.
from genesis_explorer.clients.base.base_database import BaseDatabase


class BigQueryClient(BaseDatabase):
    """Clase que define un cliente para la comunicación con BigQuery."""

    def __init__(self,
                 project_id: str) -> None:
        """Método de instanciación de la clase.
        
        Args:
        ----------
        project_id: str.
            ID del proyecto de GCP sobre el cual se creará el cliente."""
        
        logging.info("Instanciando el cliente...")
        
        self.project_id = project_id
        self.client = bigquery.Client(project = project_id)

    def create_dataset(self,
                       dataset_name: str,
                       description: Optional[str] = "",
                       labels: Optional[Dict[str, str]] = {}) -> None:
        """Método que nos permite crear un conjunto de tablas.
        
        Args:
        ----------
        dataset_name: str.
            Nombre del conjunto de tablas que se desea crear.

        description: Optional[str].
            Descripción del conjunto de tablas que se desea crear.

        labels: Optional[Dict[str, str]].
            Etiquetas del conjunto de tablas que se desea crear."""

        dataset_id = f"{self.project_id}.{dataset_name}"

        logging.info(f"Creando el conjunto de tablas: {dataset_id}.")
        
        try:
            dataset = bigquery.Dataset(dataset_id)

            dataset.labels = labels
            dataset.description = description

            dataset = self.client.create_dataset(dataset, exists_ok = True)

            logging.info(f"Conjunto de tablas creado correctamente: {dataset_id}.")
        
        except Exception as e:
            logging.error(f"Error al crear el conjunto de tablas: {e}.")

    def create_table(self,
                     table_name: str,
                     dataset_name: str,
                     schema: List[bigquery.SchemaField],
                     description: Optional[str] = "",
                     labels: Optional[Dict[str, str]] = {}) -> bigquery.Table:
        """Método que nos permite crear una tabla.
        
        Args:
        ----------
        table_name: str.
            Nombre de la tabla que se desea crear.

        dataset_name: str.
            Nombre del conjunto de tablas sobre el cual se creará la tabla.

        schema: List[bigquery.SchemaField].
            Esquema de la tabla que se desea crear.

        description: Optional[str].
            Descripción de la tabla que se desea crear.
        
        labels: Optional[Dict[str, str]].
            Etiquetas de la tabla que se desea crear.
            
        Returns:
        ----------
        bigquery.Table.
            Objeto de tipo tabla de BigQuery."""
        
        table_id = f"{self.project_id}.{dataset_name}.{table_name}"

        logging.info(f"Creando la tabla: {table_id}.")

        try:
            table = bigquery.Table(table_id, schema = schema)

            table.labels = labels
            table.description = description
            
            table = self.client.create_table(table, exists_ok = True)

            logging.info(f"Tabla creada correctamente: {table_id}.")

        except Exception as e:
            logging.error(f"Error al crear la tabla: {e}.")

        return table

    def create_table_from_dataframe(self,
                                    table_name: str,
                                    dataset_name: str,
                                    dataframe: pd.DataFrame,
                                    schema: Optional[List[bigquery.SchemaField]] = None,
                                    **kwargs) -> None:
        """Método que nos permite crear una tabla a partir de un DataFrame.
        
        Args:
        ----------
        table_name: str.
            Nombre de la tabla que se desea crear.
        
        dataset_name: str.
            Nombre del conjunto de tablas sobre el cual se creará la tabla.

        dataframe: pd.DataFrame.
            DataFrame que se desea cargar en la tabla.

        schema: Optional[List[bigquery.SchemaField]].
            Esquema de la tabla que se desea crear."""
        
        try:
            table = self.create_table(table_name, dataset_name, schema, **kwargs)

            self.client.load_table_from_dataframe(dataframe, table)

        except Exception as e:
            logging.error(f"Error al crear la tabla: {e}.")

    def get_table(self, table_name: str, dataset_name: str) -> pd.DataFrame:
        """Método que nos permite obtener una tabla.
        
        Args:
        ----------
        table_name: str.
            Nombre de la tabla que se desea obtener."""
        
        table_id = f"{self.project_id}.{dataset_name}.{table_name}"

        logging.info(f"Obteniendo la tabla: {table_id}.")
        
        try:
            table = self.client.get_table(table_id)  
            table_df = self.client.list_rows(table).to_dataframe()
            
            logging.info(f"Tabla obtenida correctamente: {table_id}.")
        
        except Exception as e:
            logging.error(f"Error al obtener la tabla: {e}.")

        return table_df
    
    def query_to_dataframe(self, query: str) -> pd.DataFrame:
        """Método que nos permite ejecutar una consulta y retornar un DataFrame.
        
        Args:
        ----------
        query: str.
            Consulta que se desea ejecutar.
            
        Returns:
        ----------
        pd.DataFrame.
            DataFrame resultante de la consulta."""
        
        logging.info(f"Ejecutando la consulta...")

        try:
            query_job = self.client.query(query)
            dataframe = query_job.to_dataframe()

            logging.info(f"Consulta ejecutada correctamente: {query_job.total_rows} filas.")
        
        except Exception as e:
            logging.error(f"Error al ejecutar la consulta: {e}.")
        
        return dataframe
    
    def export_to_storage(self,
                          table_name: str,
                          dataset_name: str,
                          bucket_name: str,
                          blob_name: str) -> None:
        """Método que nos permite exportar una tabla a un bucket de Google Cloud Storage.
        
        Args:
        ----------
        table_name: str.
            Nombre de la tabla que se desea exportar.

        dataset_name: str.
            Nombre del conjunto de tablas sobre el cual se encuentra la tabla.

        bucket_name: str.
            Nombre del bucket de Google Cloud Storage donde se desea exportar la tabla.

        blob_name: str.
            Nombre del blob de Google Cloud Storage donde se desea exportar la tabla."""
        
        table_id = f"{self.project_id}.{dataset_name}.{table_name}"
        blob_path = f"gs://{bucket_name}/{blob_name}/{table_name}_*.csv"
        
        logging.info(f"Exportando la tabla: {table_id} a {blob_path}.")
        
        try:
            job = self.client.extract_table(table_id,
                                            f"gs://{bucket_name}/{blob_name}")
            job.result()

            logging.info(f"Tabla exportada correctamente: {table_id} a {bucket_name}.")
        
        except Exception as e:
            logging.error(f"Error al exportar la tabla: {e}.")

    def __str__(self) -> str:
        """Método que retorna una representación en cadena de la clase."""

        return f"BigQueryClient(project_id = {self.project_id})"



"""Módulo que define la metaclase Singleton para nuestros clientes de GCP."""

# Librerías Externas.
from typing import Any
from abc import ABCMeta

import logging


logging.basicConfig(level = logging.INFO,
                    format = "%(asctime)s - %(levelname)s - %(message)s")


class Singleton(type):
    """Metaclase que implementa el patrón Singleton para nuestros clientes de GCP."""

    __instance = None

    def __call__(cls, *args, **kwargs) -> Any:
        """Método que intercepta el constructor e instanciador de clases
        para asegurarnos que solo se cree una instancia de la clase."""

        if cls.__instance is None:
            logging.info("Instanciando la primera instancia de la clase...")
            cls.__instance = super(Singleton, cls).__call__(*args, **kwargs)

        logging.info("Esta clase implementa el patrón Singleton, por ende, se retorna la instancia existente.")
        return cls.__instance
    

class SingletonMeta(Singleton, ABCMeta):
    """Metaclase que implementa el patrón Singleton para nuestros clientes de bases de datos."""

    ...



"""Módulo que define la clase abstracta para las Factories de servicios de nube."""

# Librerías Externas.
from typing import Dict, Type

# Librerías Internas.
from genesis_explorer.clients.base.base_storage import BaseStorage
from genesis_explorer.clients.base.base_database import BaseDatabase
from genesis_explorer.clients.base.base_functions import BaseFunction
from genesis_explorer.clients.base.base_messaging import BaseMessaging
from genesis_explorer.clients.base.base_aiplatform import BaseAIPlatform

from genesis_explorer.clients.factory.base import CloudAbstractFactory
from genesis_explorer.clients.factory.gcp_factory import GCPAbstractFactory


class CloudFactory:
    """Clase que define el contrato para la creación de Factories de servicios de nube."""

    _registry: Dict[str, Type[CloudAbstractFactory]] = {"gcp": GCPAbstractFactory, 
                                                        "aws": "AWSFactory"}

    @classmethod
    def create_database_service(cls,
                                cloud_provider: str,
                                **kwargs) -> BaseDatabase:
        """Método que retorna la Factory de servicios de nube."""

        if cloud_provider not in cls._registry:
            raise ValueError(f"El proveedor de nube {cloud_provider} no está registrado.")

        factory = cls._registry.get(cloud_provider)
        return factory.create_database_service(**kwargs)
    
    @classmethod
    def create_storage_service(cls,
                               cloud_provider: str,
                               **kwargs) -> BaseStorage:
        """Método que retorna la Factory de servicios de nube."""

        if cloud_provider not in cls._registry:
            raise ValueError(f"El proveedor de nube {cloud_provider} no está registrado.")

        factory = cls._registry.get(cloud_provider)
        return factory.create_storage_service(**kwargs)
    
    @classmethod
    def create_messaging_service(cls,
                                 cloud_provider: str,
                                 **kwargs) -> BaseMessaging:
        """Método que retorna la Factory de servicios de nube."""

        if cloud_provider not in cls._registry:
            raise ValueError(f"El proveedor de nube {cloud_provider} no está registrado.")

        factory = cls._registry.get(cloud_provider)
        return factory.create_messaging_service(**kwargs)
    
    @classmethod
    def create_function_service(cls,
                                cloud_provider: str,
                                **kwargs) -> BaseFunction:
        """Método que retorna la Factory de servicios de nube."""

        if cloud_provider not in cls._registry:
            raise ValueError(f"El proveedor de nube {cloud_provider} no está registrado.")

        factory = cls._registry.get(cloud_provider)
        return factory.create_function_service(**kwargs)
    
    @classmethod
    def create_aiplatform_service(cls,
                                cloud_provider: str,
                                **kwargs) -> BaseAIPlatform:
        """Método que retorna la Factory de servicios de nube."""

        if cloud_provider not in cls._registry:
            raise ValueError(f"El proveedor de nube {cloud_provider} no está registrado.")
        
        factory = cls._registry.get(cloud_provider)
        return factory.create_aiplatform_service(**kwargs)



"""Módulo que define un director en el patrón builder para generar flavours de validaciones."""

# Librerías Internas.
from genesis_explorer.validations.builder import Builder
from genesis_explorer.validations.product import Validator


class Director:
    """Clase que define la coordinación entre el cliente y el builder 
    para la creación de validadores."""
    
    def __init__(self, builder: Builder) -> None:
        """Método de instanciación de clase."""

        self._builder = builder
    
    def add_score_validation(self) -> None:
        """Método que define el contrato para la validación de scores."""

        self._builder.add_scores_validation()
    
    def add_metric_validation(self) -> None:
        """Método que define el contrato para la validación de métricas."""

        self._builder.add_metrics_validation()

    def add_csi_validation(self) -> None:
        """Método que define el contrato para la validación del CSI."""

        self._builder.add_csi_validation()
    
    def get_validator(self) -> Validator:
        """Método que retorna el validator."""

        return self._builder._validator


"""Implementación concreta de la clase Factory de GCS para interactura con el patrón Abstract Factory."""

# Librerías Internas.
from genesis_explorer.clients.factory.base import CloudAbstractFactory

from genesis_explorer.clients.gcp.pub_sub_client import PubSubService
from genesis_explorer.clients.gcp.aiplatform_client import AIPlatform
from genesis_explorer.clients.gcp.bigquery_client import BigQueryClient
from genesis_explorer.clients.gcp.functions_client import CloudFunction
from genesis_explorer.clients.gcp.storage_client import GoogleCloudStorageClient


class GCPAbstractFactory(CloudAbstractFactory):
    """Clase que define el contrato para la creación de Factories de GCP."""
    
    @classmethod
    def create_database_service(cls, **kwargs) -> None:
        """Método que define el contrato para la creación de servicios de bases de datos."""

        return BigQueryClient(**kwargs)
    
    @classmethod
    def create_storage_service(cls, **kwargs) -> None:
        """Método que define el contrato para la creación de servicios de almacenamiento."""

        return GoogleCloudStorageClient(**kwargs)
    
    @classmethod
    def create_messaging_service(cls, **kwargs) -> None:
        """Método que define el contrato para la creación de servicios de mensajería."""

        return PubSubService(**kwargs)
    
    @classmethod
    def create_function_service(cls, **kwargs) -> None:
        """Método que define el contrato para la creación de servicios de Cloud Functions."""

        return CloudFunction(**kwargs)
    
    @classmethod
    def create_aiplatform_service(cls, **kwargs) -> None:
        """Método que define el contrato para la creación de servicios de AI Platform."""

        return AIPlatform(**kwargs)



"""Módulo que define un builder en el patrón builder para generar flavours de validaciones."""

# Librerías Externas.
from typing import Optional
from abc import ABC, abstractmethod

import numpy as np
import pandas as pd

# Librerías Internas.
from genesis_explorer.validations.product import Validator


class Builder(ABC):
    """Clase que define el contrato estructural para creación de builders."""

    _csi_threshold: Optional[float] = 1.5
    _pvalue_threshold: Optional[float] = 0.05
    _metrics_threshold: Optional[float] = 0.05

    def __init__(self) -> None:
        """Método de instanciación de clase."""

        self._validator = Validator()
    
    @abstractmethod
    def add_scores_validation(self):
        """Método que define el contrato para la validación de scores.
        
        Returns:
        ----------
        Builder.
            Builder para la validación de scores."""
        
        raise NotImplementedError("Método no implementado.")
    
    @abstractmethod
    def add_metrics_validation(self):
        """Método que define el contrato para la validación de métricas.
        
        Returns:
        ----------
        Builder.
            Builder para la validación de métricas."""
        
        raise NotImplementedError("Método no implementado.")
    
    def add_csi_validation(self):
        """Método que, como se usa en ambos productos indiscriminadamente,
        se puede definir en el Builder."""

        def validate_csi(X_train: pd.DataFrame,
                         X_test: pd.DataFrame, **kwargs) -> bool:
            """Función que se encarga de validar el CSI del modelo.
            
            Args:
            ----------
            X_train: pd.DataFrame.
                Conjunto de datos de entrenamiento.

            X_test: pd.DataFrame.
                Conjunto de datos de prueba.
            
            Returns:
            ----------
            bool.
                True si el CSI del modelo es válido, False en caso contrario."""

            try:
                if hasattr(X_train, "values"):
                    X_train = X_train.values
                if hasattr(X_test, "values"):
                    X_test = X_test.values
            except AttributeError as e:
                raise AttributeError(f"Error al validar el CSI del modelo: {e}")
            
            X_train_mean = np.mean(X_train, axis = 0)
            X_test_mean = np.mean(X_test, axis = 0)

            X_train_std = np.std(X_train, axis = 0)
            X_test_std = np.std(X_test, axis = 0)

            X_train_idx = X_train_mean / (X_train_std + 1e-10)
            X_test_idx = X_test_mean / (X_test_std + 1e-10)

            return all(np.abs(X_train_idx - X_test_idx) < self._csi_threshold)

        self._validator.add_validation(validate_csi)
        return self



"""Módulo que contiene la clase AIPlatform para manejar el tracking de experimentos y runs en Vertex AI."""

# Librerías Externas.
from typing import Any, TypeVar, List, Optional

import os
import json
import pickle
import logging

from google.cloud import aiplatform
from google.api_core.exceptions import AlreadyExists, NotFound

# Librerías Internas.
from genesis_explorer.clients.base.base_aiplatform import BaseAIPlatform


logging.basicConfig(level = logging.INFO,
                    format = "%(asctime)s - %(name)s - %(levelname)s - %(message)s")


Model = TypeVar("Model")
Experiment = TypeVar("Experiment")
ExperimentRun = TypeVar("ExperimentRun")


class AIPlatform(BaseAIPlatform):
    """Clase que se encarga de manejar el tracking de experimentos y runs en Vertex AI."""

    def __init__(self, project_id: str,
                 location: Optional[str] = "us-east1",
                 staging_bucket: Optional[str] = None) -> None:
        """Método de instanciación de clase.
        
        Args:
        ----------
        project_id: str.
            ID del proyecto de GCP.
        
        location: Optional[str].
            Región de GCP donde guardará registros.
        
        staging_bucket: Optional[str].
            Bucket de GCS donde se guardarán los archivos temporales."""
        
        aiplatform.init(project = project_id,
                        location = location,
                        staging_bucket = staging_bucket)
        
    def create_experiment(self, experiment_name: str) -> Experiment:
        """Método que permite crear u obtener (si ya existe) un experimento.
        
        Args:
        ----------
        experiment_name: str.
            Nombre del experimento.
        
        Returns:
        ----------
        Experiment.
            Experimento a usar para registrar ExperimentRuns."""

        logging.info(f"Creando experimento: {experiment_name}")

        try:
            experiment = aiplatform.Experiment.create(experiment_name = experiment_name)

            logging.info(f"Experimento creado: {experiment_name}")
            return experiment
        except AlreadyExists as e:
            logging.info(f"Experimento ya existe: {experiment_name}")
            experiment = aiplatform.Experiment.get(experiment_name = experiment_name)
            return experiment
        except Exception as e:
            logging.error(f"Error al crear experimento: {e}")
            raise e
        
    def get_experiment(self, experiment_name: str) -> Experiment:
        """Método que permite obtener un experimento ya registrado para
        actualizarlo o, simplemente, leerlo.
        
        Args:
        ----------
        experiment_name: str.
            Nombre del experimento."""
        
        logging.info(f"Obteniendo experimento: {experiment_name}")

        try:
            experiment = aiplatform.Experiment.get(experiment_name = experiment_name)
            logging.info(f"Experimento obtenido: {experiment_name}")
            return experiment
        except Exception as e:
            logging.error(f"Error al obtener experimento: {e}")
            raise e
        
    def delete_experiment(self, experiment_name: str) -> None:
        """Método que permite eliminar un experimento.
        
        Args:
        ----------
        experiment_name: str.
            Nombre del experimento."""
        
        logging.info(f"Eliminando experimento: {experiment_name}")

        experiment = self.create_experiment(experiment_name = experiment_name)
        try:
            experiment.delete()
            logging.info(f"Experimento eliminado: {experiment_name}")
        except Exception as e:
            logging.error(f"Error al eliminar experimento: {e}")
            raise e
        
    def list_experiments(self) -> List[Experiment]:
        """Método que permite obtener todos los experimentos en la ubicación de su AIPlatform.

        (Este método también lo podríamos aplicar como df = experiment.get_data_frame()
        para tener descripciones sin necesidad de recorrer Experiments).
        
        Returns:
        ----------
        experiments: List[Experiment].
            Lista de experimentos."""
        
        logging.info("Obteniendo experimentos...")
        try:
            experiments = aiplatform.Experiment.list()
            logging.info(f"Experimentos obtenidos: {experiments}")
            return experiments
        except Exception as e:
            logging.error(f"Error al obtener experimentos: {e}")
            raise e
        
    def create_run(self, run_name: str, experiment: Experiment) -> ExperimentRun:
        """Método que permite crear un ExperimentRun.
        
        Args:
        ----------
        run_name: str.
            Nombre del ExperimentRun.
        
        experiment: Experiment.
            Experimento al que se asociará el ExperimentRun.
        
        Returns:
        ----------
        ExperimentRun.
            ExperimentRun creado para registrar su experimento particular."""
        
        logging.info(f"Creando ExperimentRun: {run_name}")

        try:
            run = aiplatform.ExperimentRun.create(run_name = run_name, experiment = experiment)
            logging.info(f"ExperimentRun creado: {run_name}")
            return run
        except Exception as e:
            logging.error(f"Error al crear ExperimentRun: {e}")
            raise e
        
    def get_run(self, run_name: str, experiment: Experiment) -> ExperimentRun:
        """Método que permite obtener un ExperimentRun ya registrado para
        actualizarlo o, simplemente, leerlo.
        
        Args:
        ----------
        run_name: str.
            Nombre del ExperimentRun.
        
        experiment: aiplatform.metadata.experiment_resources.Experiment.
            Experimento al que se asoció el ExperimentRun.
        
        Returns:
        ----------
        run: aiplatform.metadata.experiment_resources.ExperimentRun.
            ExperimentRun obtenido para actualizar o leer."""
        
        logging.info(f"Obteniendo ExperimentRun: {run_name}")
        
        try:
            run = aiplatform.ExperimentRun.get(run_name = run_name, experiment = experiment)
            logging.info(f"ExperimentRun obtenido: {run_name}")
            return run
        except Exception as e:
            logging.error(f"Error al obtener ExperimentRun: {e}")
            raise e

    def delete_run(self, experiment_run: ExperimentRun) -> None:
        """Método que permite eliminar un ExperimentRun.
        
        Args:
        ----------
        experiment_run: ExperimentRun.
            Instancia del ExperimentRun a eliminar."""
        
        logging.info(f"Eliminando ExperimentRun: {experiment_run.name}")

        try:
            experiment_run.delete()
            logging.info(f"ExperimentRun eliminado: {experiment_run.name}")
        except Exception as e:
            logging.error(f"Error al eliminar ExperimentRun: {e}")
            raise e
        
    def end_run(self, experiment_run: ExperimentRun) -> None:
        """Método que permite cerrar un ExperimentRun.
        
        Args:
        ----------
        experiment_run: ExperimentRun.
            Instancia del ExperimentRun a cerrar."""
        
        logging.info(f"Cerrando ExperimentRun: {experiment_run.name}")

        try:
            experiment_run.end_run()
            logging.info(f"ExperimentRun cerrado: {experiment_run.name}")
        except Exception as e:
            logging.error(f"Error al cerrar ExperimentRun: {e}")
            raise e
    
    def list_runs(self, experiment: Experiment) -> List[ExperimentRun]:
        """Método que permite obtener todos los ExperimentRuns asociados a un Experiment.
        
        Args:
        ----------
        experiment: Experiment.
            Experimento al que se asociaron los ExperimentRuns.
        
        Returns:
        ----------
        experiment_runs: List[ExperimentRun].
            Lista de ExperimentRuns asociados al Experiment."""

        logging.info(f"Obteniendo ExperimentRuns asociados al Experiment: {experiment.name}")

        try:
            experiment_runs = aiplatform.ExperimentRun.list(experiment = experiment)
            logging.info(f"ExperimentRuns obtenidos: {experiment_runs}")
            return experiment_runs
        except Exception as e:
            logging.error(f"Error al obtener ExperimentRuns: {e}")
            raise e
        
    def create_tabular_dataset(self, name: str,
                               gcs_source: Optional[str] = None, bq_source: Optional[str] = None) -> aiplatform.TabularDataset:
        """Método que permite crear un Dataset dentro de Vertex AI
        para dar un punto de inicio a la experimentación.

        Esto se guarda en las pestañas Dataset de Vertex AI.
        
        Args:
        ----------
        name: str.
            Nombre del TabularDataset.
        
        gcs_source: Optional[str].
            URI de GCS donde se encuentra el dataset para referencia. (Ya debe estar creado en GCS).
        
        bq_source: Optional[str].
            URI de BigQuery donde se encuentra el dataset para referencia. (Ya debe estar creado en BigQuery).
        
        Returns:
        ----------
        dataset: aiplatform.TabularDataset.
            Dataset creado para dar un punto de inicio a la experimentación."""

        logging.info(f"Creando TabularDataset: {name}")

        try:
            dataset = aiplatform.TabularDataset.create(display_name = name,
                                                       gcs_source = gcs_source,
                                                       bq_source = bq_source)
            logging.info(f"TabularDataset creado: {name}")
            return dataset
        except Exception as e:
            logging.error(f"Error al crear TabularDataset: {e}")
            raise e
        
    def upload_trained_model(self, name: str, gcs_source: str, serving_container_image_uri: str) -> aiplatform.Model:
        """Método para cargar un modelo entrenado en GCS a Vertex AI.
        
        Args:
        ----------
        name: str.
            Nombre del modelo.
        
        gcs_source: str.
            URI de GCS donde se encuentra el modelo entrenado.
            (Ya debe estar creado en GCS y debe tener formato "gs://my-bucket/models/mlb/v1/", es importante que lo guarde como "model.pkl").
        
        serving_container_image_uri: str.
            URI del contenedor de Vertex AI que contiene referencia al modelo entrenado.
            Usualmente se usa "us-docker.pkg.dev/vertex-ai/prediction/sklearn-cpu.1-3:latest".
        
        Returns:
        ----------
        model: aiplatform.Model.
            Modelo cargado en Vertex AI."""

        logging.info(f"Cargando modelo entrenado en Vertex AI: {name}")

        try:
            model = aiplatform.Model.upload(display_name = name,
                                            parent_model = name,
                                            artifact_uri = gcs_source,
                                            serving_container_image_uri = serving_container_image_uri)
            logging.info(f"Modelo cargado: {name}")
            return model
        except NotFound as e:
            logging.error(f"Parent Model no encontrado: {name}. Creando versión 1.")
            model = aiplatform.Model.upload(display_name = name,
                                            parent_model = None,
                                            artifact_uri = gcs_source,
                                            serving_container_image_uri = serving_container_image_uri)
            logging.info(f"Modelo cargado: {name}")
            return model
        except Exception as e:
            logging.error(f"Error al cargar modelo: {e}")
            raise e

    def get_trained_model(self, name: str) -> aiplatform.Model:
        """Método para obtener un modelo entrenado en Vertex AI.
        
        Args:
        ----------
        name: str.
            Nombre del modelo."""
        
        logging.info(f"Obteniendo modelo entrenado en Vertex AI: {name}")

        try:
            model = aiplatform.Model(model_name = name)
            logging.info(f"Modelo obtenido: {name}")
            return model
        except Exception as e:
            logging.error(f"Error al obtener modelo: {e}")
            raise e
        
    @staticmethod
    def save_ephemeral_file(vertex_object: Any, content: Any, file_name: str) -> None:
        """Método estático para comunicación entre componentes de KubeFlow.
        
        Args:
        ----------
        vertex_object: Any.
            Objeto de comunicación entre Components.
            
        content: Any.
            Contenido del archivo.
        
        file_name: str.
            Nombre del archivo."""

        name, extension = file_name.split(".")

        os.makedirs(vertex_object.path, exist_ok = True)
        path = os.path.join(vertex_object.path, file_name)

        if extension == "csv":
            content.to_csv(path, index = False)
        elif extension == "parquet":
            content.to_parquet(path, index = False)
        elif extension == "json":
            with open(path, "w") as f:
                json.dump(content, f)
        elif extension == "pickle":
            with open(path, "wb") as f:
                pickle.dump(content, f)
        else:
            raise ValueError(f"Extensión {extension} no soportada.")
            



"""Módulo que contiene la clase AIPlatform para manejar el tracking de experimentos y runs en Vertex AI."""

# Librerías Externas.
from typing import TypeVar, List, Optional
from abc import ABC, abstractmethod

# Librerías Internas.
from genesis_explorer.clients.base.core.singleton import SingletonMeta


Model = TypeVar("Model")
Experiment = TypeVar("Experiment")
ExperimentRun = TypeVar("ExperimentRun")
TabularDataset = TypeVar("TabularDataset")


class BaseAIPlatform(ABC, metaclass = SingletonMeta):
    """Clase que se encarga de manejar el tracking de experimentos y runs en Vertex AI."""

    @abstractmethod
    def create_experiment(self, experiment_name: str) -> Experiment:
        """Método abstractoque permite crear u obtener (si ya existe) un experimento.
        
        Args:
        ----------
        experiment_name: str.
            Nombre del experimento.
        
        Returns:
        ----------
        Experiment.
            Experimento a usar para registrar ExperimentRuns."""

        raise NotImplementedError("Si tu clase implementa el contrato BaseAIPlatform, debes implementar el método 'create_experiment'.")
    
    def delete_experiment(self, experiment_name: str) -> None:
        """Método que permite eliminar un experimento.
        
        Args:
        ----------
        experiment_name: str.
            Nombre del experimento."""
        
        raise NotImplementedError("Si tu clase implementa el contrato BaseAIPlatform, debes implementar el método 'delete_experiment'.")
        
    def list_experiments(self) -> List[Experiment]:
        """Método que permite obtener todos los experimentos en la ubicación de su AIPlatform.

        (Este método también lo podríamos aplicar como df = experiment.get_data_frame()
        para tener descripciones sin necesidad de recorrer Experiments).
        
        Returns:
        ----------
        experiments: List[Experiment].
            Lista de experimentos."""
        
        raise NotImplementedError("Si tu clase implementa el contrato BaseAIPlatform, debes implementar el método 'list_experiments'.")
        
    def create_run(self, run_name: str, experiment: Experiment) -> ExperimentRun:
        """Método que permite crear un ExperimentRun.
        
        Args:
        ----------
        run_name: str.
            Nombre del ExperimentRun.
        
        experiment: Experiment.
            Experimento al que se asociará el ExperimentRun.
        
        Returns:
        ----------
        ExperimentRun.
            ExperimentRun creado para registrar su experimento particular."""
        
        raise NotImplementedError("Si tu clase implementa el contrato BaseAIPlatform, debes implementar el método 'create_run'.")
        
    def get_run(self, run_name: str, experiment: Experiment) -> ExperimentRun:
        """Método que permite obtener un ExperimentRun ya registrado para
        actualizarlo o, simplemente, leerlo.
        
        Args:
        ----------
        run_name: str.
            Nombre del ExperimentRun.
        
        experiment: aiplatform.metadata.experiment_resources.Experiment.
            Experimento al que se asoció el ExperimentRun.
        
        Returns:
        ----------
        run: aiplatform.metadata.experiment_resources.ExperimentRun.
            ExperimentRun obtenido para actualizar o leer."""
        
        raise NotImplementedError("Si tu clase implementa el contrato BaseAIPlatform, debes implementar el método 'get_run'.")
    
    def delete_run(self, experiment_run: ExperimentRun) -> None:
        """Método que permite eliminar un ExperimentRun.
        
        Args:
        ----------
        experiment_run: ExperimentRun.
            Instancia del ExperimentRun a eliminar."""
        
        raise NotImplementedError("Si tu clase implementa el contrato BaseAIPlatform, debes implementar el método 'delete_run'.")
        
    def end_run(self, experiment_run: ExperimentRun) -> None:
        """Método que permite cerrar un ExperimentRun.
        
        Args:
        ----------
        experiment_run: ExperimentRun.
            Instancia del ExperimentRun a cerrar."""
        
        raise NotImplementedError("Si tu clase implementa el contrato BaseAIPlatform, debes implementar el método 'end_run'.")
    
    def list_runs(self, experiment: Experiment) -> List[ExperimentRun]:
        """Método que permite obtener todos los ExperimentRuns asociados a un Experiment.
        
        Args:
        ----------
        experiment: Experiment.
            Experimento al que se asociaron los ExperimentRuns.
        
        Returns:
        ----------
        experiment_runs: List[ExperimentRun].
            Lista de ExperimentRuns asociados al Experiment."""

        raise NotImplementedError("Si tu clase implementa el contrato BaseAIPlatform, debes implementar el método 'list_runs'.")
        
    def create_tabular_dataset(self, name: str,
                               gcs_source: Optional[str] = None, bq_source: Optional[str] = None) -> TabularDataset:
        """Método que permite crear un Dataset dentro de Vertex AI
        para dar un punto de inicio a la experimentación.

        Esto se guarda en las pestañas Dataset de Vertex AI.
        
        Args:
        ----------
        name: str.
            Nombre del TabularDataset.
        
        gcs_source: Optional[str].
            URI de storage donde se encuentra el dataset para referencia. (Ya debe estar creado, por ejemplo en GCS).
        
        bq_source: Optional[str].
            URI de database donde se encuentra el dataset para referencia. (Ya debe estar creado, por ejemplo en BigQuery).
        
        Returns:
        ----------
        dataset: aiplatform.TabularDataset.
            Dataset creado para dar un punto de inicio a la experimentación."""

        raise NotImplementedError("Si tu clase implementa el contrato BaseAIPlatform, debes implementar el método 'create_tabular_dataset'.")
        
    def upload_trained_model(self, name: str, gcs_source: str, serving_container_image_uri: str) -> Model:
        """Método para cargar un modelo entrenado la plataforma.
        
        Args:
        ----------
        name: str.
            Nombre del modelo.
        
        gcs_source: str.
            URI de storage donde se encuentra el modelo entrenado.
        
        serving_container_image_uri: str.
            URI del contenedor de que contiene referencia al modelo entrenado.
        
        Returns:
        ----------
        model: aiplatform.Model.
            Modelo cargado en Vertex AI."""

        raise NotImplementedError("Si tu clase implementa el contrato BaseAIPlatform, debes implementar el método 'upload_trained_model'.")

    def get_trained_model(self, name: str) -> Model:
        """Método para obtener un modelo entrenado en Vertex AI.
        
        Args:
        ----------
        name: str.
            Nombre del modelo."""
        
        raise NotImplementedError("Si tu clase implementa el contrato BaseAIPlatform, debes implementar el método 'get_trained_model'.")



"""Módulo que define el contrato para los clientes de servicios de Cloud Functions."""

# Librerías Externas.
from abc import ABC, abstractmethod

# Librerías Internas.
from genesis_explorer.clients.base.core.singleton import SingletonMeta


class BaseFunction(ABC, metaclass = SingletonMeta):
    """Clase que define un cliente para la comunicación con Cloud Functions."""
    
    @abstractmethod
    def create_function(self, **kwargs) -> None:
        """Método que, dada la ruta a su archivo zip (que contiene el script de la función
        y sus requirements) en GCS, crea una Cloud Function en GCP."""

        raise NotImplementedError("Si tu clase implementa el contrato BaseFunction, debes implementar el método 'create_function'.")



"""Módulo que define la clase base para los clientes de bases de datos."""

# Librerías Externas.
from typing import Any
from abc import ABC, abstractmethod

import pandas as pd

# Librerías Internas.
from genesis_explorer.clients.base.core.singleton import SingletonMeta


class BaseStorage(ABC, metaclass = SingletonMeta):
    """Clase que implementa el contrato para los clientes de servicios de storage."""
    
    @abstractmethod
    def create_bucket(self,
                      bucket_name: str) -> None:
        """Método que nos permite crear un bucket en Google Cloud Storage.
        
        Args:
        ----------
        bucket_name: str.
            Nombre del bucket que se desea crear."""

        raise NotImplementedError("Si tu clase implementa el contrato BaseStorage, debes implementar el método 'create_bucket'.")
    
    @abstractmethod
    def upload_file(self,
                    bucket_name: str,
                    file_path: str,
                    destination_path: str) -> None:
        """Método que nos permite subir un archivo a Google Cloud Storage.

        Args:
        ----------
        bucket_name: str.
            Nombre del bucket en el cual se subirá el archivo.

        file_path: str.
            Ruta del archivo que se desea subir.

        destination_path: str.
            Ruta en la cual se subirá el archivo. """

        raise NotImplementedError("Si tu clase implementa el contrato BaseStorage, debes implementar el método 'upload_file'.")

    @abstractmethod
    def download_file(self,
                      bucket_name: str,
                      file_path: str,
                      destination_path: str) -> None:
        """Método que nos permite descargar un archivo de Google Cloud Storage.

        Args:
        ----------
        bucket_name: str.
            Nombre del bucket en el cual se encuentra el archivo.

        file_path: str.
            Ruta del archivo que se desea descargar.

        destination_path: str.
            Ruta en la cual se descargará el archivo. """

        raise NotImplementedError("Si tu clase implementa el contrato BaseStorage, debes implementar el método 'download_file'.")
    
    @abstractmethod
    def download_csv(self,
                     bucket_name: str,
                     file_path: str) -> pd.DataFrame:
        """Método que nos permite descargar un archivo CSV de Google Cloud Storage.
        
        Args:
        ----------
        bucket_name: str.
            Nombre del bucket en el cual se encuentra el archivo.
        
        file_path: str.
            Ruta del archivo que se desea descargar.
            
        Returns:
        ----------
        df: pd.DataFrame.
            DataFrame con los datos del archivo CSV. """

        raise NotImplementedError("Si tu clase implementa el contrato BaseStorage, debes implementar el método 'download_csv'.")
    
    @abstractmethod
    def download_pickle(self,
                        bucket_name: str,
                        file_path: str) -> Any:
        """Método que nos permite descargar un archivo pickle de Google Cloud Storage.
        
        Args:
        ----------
        bucket_name: str.
            Nombre del bucket en el cual se encuentra el archivo.
        
        file_path: str.
            Ruta del archivo que se desea descargar.

        Returns:
        ----------
        Any.
            Objeto deserializado desde el archivo pickle."""
        
        raise NotImplementedError("Si tu clase implementa el contrato BaseStorage, debes implementar el método 'download_pickle'.")



"""Módulo que define un cliente para la comunicación con Google Cloud Storage."""

# Librerías Externas.
from typing import Any

import pickle
import logging
from io import StringIO

import pandas as pd

from google.cloud import storage

# Librerías Internas.
from genesis_explorer.clients.base.base_storage import BaseStorage


logging.basicConfig(level = logging.INFO,
                    format = "%(asctime)s - %(levelname)s - %(message)s")


class GoogleCloudStorageClient(BaseStorage):
    """Clase que define un cliente para la comunicación con Google Cloud Storage."""

    def __init__(self,
                 project_id: str) -> None:
        """Método de instanciación de la clase.
        
        Args:
        ----------
        project_id: str.
            ID del proyecto de GCP sobre el cual se creará el cliente."""
        
        self.project_id = project_id
        self.client = storage.Client(project = project_id)

    def create_bucket(self,
                      bucket_name: str) -> None:
        """Método que nos permite crear un bucket en Google Cloud Storage.
        
        Args:
        ----------
        bucket_name: str.
            Nombre del bucket que se desea crear."""

        logging.info(f"Creando el bucket {bucket_name} en el proyecto {self.project_id}...")
        
        try:
            bucket = self.client.create_bucket(bucket_name)
            logging.info(f"Bucket {bucket_name} creado correctamente.")
        except Exception as e:
            logging.error(f"Error al crear el bucket {bucket_name}: {e}")

        return bucket
        
    def upload_file(self,
                    bucket_name: str,
                    file_path: str,
                    destination_path: str) -> None:
        """Método que nos permite subir un archivo a Google Cloud Storage.

        Args:
        ----------
        bucket_name: str.
            Nombre del bucket en el cual se subirá el archivo.

        file_path: str.
            Ruta del archivo que se desea subir.

        destination_path: str.
            Ruta en la cual se subirá el archivo. """

        logging.info(f"Subiendo el archivo {file_path} al bucket {bucket_name} en el proyecto {self.project_id}...")

        try:
            bucket = self.client.lookup_bucket(bucket_name)
            if not bucket:
                bucket = self.create_bucket(bucket_name)
            
            blob = bucket.blob(destination_path)
            blob.upload_from_filename(file_path)
            logging.info(f"Archivo {file_path} subido correctamente.")

        except Exception as e:
            logging.error(f"Error al subir el archivo {file_path}: {e}")

    def download_file(self,
                      bucket_name: str,
                      file_path: str,
                      destination_path: str) -> None:
        """Método que nos permite descargar un archivo de Google Cloud Storage.

        Args:
        ----------
        bucket_name: str.
            Nombre del bucket en el cual se encuentra el archivo.

        file_path: str.
            Ruta del archivo que se desea descargar.

        destination_path: str.
            Ruta en la cual se descargará el archivo. """

        logging.info(f"Descargando el archivo {file_path} del bucket {bucket_name} en el proyecto {self.project_id}...")

        try:
            bucket = self.client.lookup_bucket(bucket_name)
            if not bucket:
                bucket = self.create_bucket(bucket_name)

            blob = bucket.blob(file_path)
            blob.download_to_filename(destination_path)
            logging.info(f"Archivo {file_path} descargado correctamente.")

        except Exception as e:
            logging.error(f"Error al descargar el archivo {file_path}: {e}")

    def download_csv(self,
                     bucket_name: str,
                     file_path: str) -> pd.DataFrame:
        """Método que nos permite descargar un archivo CSV de Google Cloud Storage.
        
        Args:
        ----------
        bucket_name: str.
            Nombre del bucket en el cual se encuentra el archivo.

        file_path: str.
            Ruta del archivo que se desea descargar.
            
        Returns:
        ----------
        df: pd.DataFrame.
            DataFrame con los datos del archivo CSV. """
        
        logging.info(f"Descargando dataset del bucket {bucket_name} en el proyecto {self.project_id}...")
        
        try:
            bucket = self.client.lookup_bucket(bucket_name)
            if not bucket:
                bucket = self.create_bucket(bucket_name)

            blob = bucket.blob(file_path)

            data = blob.download_as_text()
            df = pd.read_csv(StringIO(data))
            logging.info(f"Archivo {file_path} descargado correctamente.")

        except Exception as e:
            logging.error(f"Error al descargar el archivo {file_path}: {e}")

        return df
    
    def download_pickle(self,
                        bucket_name: str,
                        file_path: str) -> Any:
        """Método que nos permite descargar un archivo pickle de Google Cloud Storage.
        
        Args:
        ----------
        bucket_name: str.
            Nombre del bucket en el cual se encuentra el archivo.

        file_path: str.
            Ruta del archivo que se desea descargar.

        Returns:
        ----------
        Any.
            Objeto deserializado desde el archivo pickle."""
        
        logging.info(f"Descargando archivo pickle del bucket {bucket_name} en el proyecto {self.project_id}...")
        
        try:
            bucket = self.client.lookup_bucket(bucket_name)
            if not bucket:
                bucket = self.create_bucket(bucket_name)
                
            blob = bucket.blob(file_path)
            data = blob.download_as_bytes()
            modelo = pickle.loads(data)

            logging.info(f"Archivo {file_path} descargado correctamente.")

        except Exception as e:
            logging.error(f"Error al descargar el archivo {file_path}: {e}")

        return modelo

    def __str__(self) -> str:
        return f"GoogleCloudStorageClient(project_id = {self.project_id})"



"""Módulo que define la clase base para las Factories de servicios de nube."""

# Librerías Externas.
from abc import ABC, abstractmethod


class CloudAbstractFactory(ABC):
    """Clase que define el contrato para la creación de Factories."""

    @abstractmethod
    def create_database_service(self,
                                *args, **kwargs) -> None:
        """Método que define el contrato para la creación de servicios de bases de datos."""

        raise NotImplementedError("Si tu clase implementa el contrato CloudFactory, debes implementar el método 'create_database_service'.")
    
    @abstractmethod
    def create_storage_service(self, *args, **kwargs) -> None:
        """Método que define el contrato para la creación de servicios de almacenamiento."""

        raise NotImplementedError("Si tu clase implementa el contrato CloudFactory, debes implementar el método 'create_storage_service'.")
    
    @abstractmethod
    def create_messaging_service(self, *args, **kwargs) -> None:
        """Método que define el contrato para la creación de servicios de mensajería."""

        raise NotImplementedError("Si tu clase implementa el contrato CloudFactory, debes implementar el método 'create_messaging_service'.")
    
    @abstractmethod
    def create_function_service(self, *args, **kwargs) -> None:
        """Método que define el contrato para la creación de servicios de Cloud Functions."""

        raise NotImplementedError("Si tu clase implementa el contrato CloudFactory, debes implementar el método 'create_function_service'.")
    
    @abstractmethod
    def create_aiplatform_service(self, *args, **kwargs) -> None:
        """Método que define el contrato para la creación de servicios de AI Platform."""

        raise NotImplementedError("Si tu clase implementa el contrato CloudFactory, debes implementar el método 'create_aiplatform_service'.")



"""Módulo que define la clase base para los clientes de mensajería."""

# Librerías Externas.
from typing import Callable, Optional, List, Dict, Any
from abc import ABC, abstractmethod

# Librerías Internas.
from genesis_explorer.clients.base.core.singleton import SingletonMeta


class BaseMessaging(ABC, metaclass = SingletonMeta):
    """Clase que implementa el contrato de servicios de mensajería."""

    @abstractmethod
    def create_topic(self, topic_name: str) -> None:
        """Método para crear un tópico.
        
        Args:
        ----------
        topic_name: str
            Nombre del tópico."""
        
        raise NotImplementedError("Si tu clase implementa el contrato BaseMessaging, debes implementar el método 'create_topic'.")

    @abstractmethod
    def publish(self, topic_name: str, message: str) -> None:
        """Método para publicar un mensaje en un tópico.
        
        Args:
        ----------
        topic_name: str
            Nombre del tópico.

        message: str
            Mensaje a publicar."""
        
        raise NotImplementedError("Si tu clase implementa el contrato BaseMessaging, debes implementar el método 'publish'.")

    @abstractmethod
    def batch_pull(self, sub_name: str, max_messages: Optional[int] = 10) -> List[Dict[str, Any]]:
        """Método para obtener mensajes de un tópico.
        
        Args:
        ----------
        sub_name: str
            Nombre de la suscripción.

        max_messages: Optional[int]
            Número máximo de mensajes a obtener.

        Returns:
        ----------
        List[Dict[str, Any]].
            Lista de mensajes obtenidos."""
        
        raise NotImplementedError("Si tu clase implementa el contrato BaseMessaging, debes implementar el método 'pull'.")

    @abstractmethod
    def subscribe(self, sub_name: str, topic_name: str) -> None:
        """Método para suscribirse a un tópico.
        
        Args:
        ----------
        sub_name: str
            Nombre de la suscripción.

        topic_name: str
            Nombre del tópico."""
        
        raise NotImplementedError("Si tu clase implementa el contrato BaseMessaging, debes implementar el método 'subscribe'.")
    
    @abstractmethod
    def stream_pull(self, sub_name: str, callback: Callable) -> Any:
        """Método para suscribirse a un tópico en streaming.
        
        Args:
        ----------
        sub_name: str
            Nombre de la suscripción.
            
        callback: Callable
            Función de callback para procesar los mensajes.

        Returns:
        ----------
        Any.
            Cliente de suscripción."""
        
        raise NotImplementedError("Si tu clase implementa el contrato BaseMessaging, debes implementar el método 'stream_pull'.")



"""Módulo que define la clase base para los clientes de bases de datos."""

# Librerías Externas.
from abc import ABC, abstractmethod

import pandas as pd

# Librerías Internas.
from genesis_explorer.clients.base.core.singleton import SingletonMeta


class BaseDatabase(ABC, metaclass = SingletonMeta):
    """Clase que implementa el contrato para los clientes de bases de datos."""

    @abstractmethod
    def create_dataset(self,
                       dataset_name: str,
                       *args, **kwargs) -> None:
        """Método que nos permite crear un conjunto de tablas.
        
        Args:
        ----------
        dataset_name: str.
            Nombre del conjunto de tablas que se desea crear."""

        raise NotImplementedError("Si tu clase implementa el contrato BaseDatabase, debes implementar el método 'create_dataset'.")

    @abstractmethod
    def create_table(self,
                     table_name: str,
                     dataset_name: str,
                     *args, **kwargs) -> None:
        """Método que nos permite crear una tabla.
        
        Args:
        ----------
        table_name: str.
            Nombre de la tabla que se desea crear.

        dataset_name: str.
            Nombre del conjunto de tablas sobre el cual se creará la tabla."""
        
        raise NotImplementedError("Si tu clase implementa el contrato BaseDatabase, debes implementar el método 'create_table'.")

    @abstractmethod
    def create_table_from_dataframe(self,
                                    table_name: str,
                                    dataset_name: str,
                                    dataframe: pd.DataFrame,
                                    *args, **kwargs) -> None:
        """Método que nos permite crear una tabla a partir de un DataFrame.
        
        Args:
        ----------
        table_name: str.
            Nombre de la tabla que se desea crear.
        
        dataset_name: str.
            Nombre del conjunto de tablas sobre el cual se creará la tabla.

        dataframe: pd.DataFrame.
            DataFrame que se desea cargar en la tabla.."""
        
        raise NotImplementedError("Si tu clase implementa el contrato BaseDatabase, debes implementar el método 'create_table_from_dataframe'.")

    @abstractmethod
    def get_table(self,
                  table_name: str,
                  dataset_name: str,
                  *args, **kwargs) -> None:
        """Método que nos permite obtener una tabla.
        
        Args:
        ----------
        table_name: str.
            Nombre de la tabla que se desea obtener."""
        
        raise NotImplementedError("Si tu clase implementa el contrato BaseDatabase, debes implementar el método 'get_table'.")
    
    @abstractmethod
    def query_to_dataframe(self,
                           query: str,
                           *args, **kwargs) -> None:
        """Método que nos permite ejecutar una consulta y retornar un DataFrame.
        
        Args:
        ----------
        query: str.
            Consulta que se desea ejecutar.
            
        Returns:
        ----------
        pd.DataFrame.
            DataFrame resultante de la consulta."""
        
        raise NotImplementedError("Si tu clase implementa el contrato BaseDatabase, debes implementar el método 'query_to_dataframe'.")
    
    @abstractmethod
    def export_to_storage(self,
                          table_name: str,
                          dataset_name: str,
                          bucket_name: str,
                          blob_name: str) -> None:
        """Método que nos permite exportar una tabla a un servicio de storage.
        
        Args:
        ----------
        table_name: str.
            Nombre de la tabla que se desea exportar.

        dataset_name: str.
            Nombre del conjunto de tablas sobre el cual se encuentra la tabla.

        bucket_name: str.
            Nombre del bucket de Google Cloud Storage donde se desea exportar la tabla.

        blob_name: str.
            Nombre del blob de Google Cloud Storage donde se desea exportar la tabla."""
        
        raise NotImplementedError("Si tu clase implementa el contrato BaseDatabase, debes implementar el método 'export_to_storage'.")
    


"""Clase que implementa el patrón Registry para manejar archivos locales."""

# Librerías Externas.
from typing import Type

# Librerías Internas.
from genesis_explorer.io.base import LocalFileBase

from genesis_explorer.io.csv_file import CSVFile
from genesis_explorer.io.json_file import JSONFile


class LocalFileRegistry:
    """Clase que implementa el patrón Registry para manejar archivos locales."""

    _registry: dict[str, Type[LocalFileBase]] = {"csv": CSVFile,
                                                 "json": JSONFile}
    
    @classmethod
    def get_file_handler(cls, file_type: str) -> Type[LocalFileBase]:
        """Método para obtener un archivo local."""

        if file_type not in cls._registry:
            raise ValueError(f"Tipo de archivo no soportado: {file_type}")

        return cls._registry[file_type]
    
    @classmethod
    def register_file_handler(cls, file_type: str) -> None:
        """Método para registrar un archivo local.
        
        Args:
        ----------
        file_type: str.
            Tipo de manejo de archivo que se desea registrar."""

        def wrapper(cls_to_register: type[LocalFileBase]) -> None:
            """Wraper que representa una clase que implementa el contrato LocalFileBase."""

            if file_type in cls._registry:
                raise ValueError(f"Tipo de archivo ya registrado: {file_type}")
            
            cls._registry[file_type] = cls_to_register
        
        return wrapper



"""Módulo que, en el patrón builder, define el producto final de la validación."""

# Librerías Externas.
from typing import Dict, Callable

import logging


class Validator:
    """Clase que define un objeto complejo usado para validar modelos de Machine Learning."""

    def __init__(self) -> None:
        """Método de instanciación de clase."""

        self.validations = []

    def add_validation(self, validation_func: Callable[..., bool]) -> None:
        """Método que permite ir creando el objeto complejo.
        
        Args:
        ----------
        validation_func: Callable[..., bool].
            Función que define la validación a realizar."""

        self.validations.append(validation_func)

    def execute_validations(self, *args, **kwargs) -> Dict[str, bool]:
        """Método que ejecuta las validaciones.
        
        Returns:
        ----------
        Dict[str, bool].
            Diccionario con los resultados de las validaciones."""

        results = {}
        for func in self.validations:
            func_name = func.__name__
            try:
                results[func_name] = func(*args, **kwargs)
            except Exception as e:
                results[func_name] = f"Error: {e}"
                logging.error(f"Error al ejecutar la validación {func_name}: {e}")
            finally:
                logging.info(f"✅ Validación {func_name} completada correctamente.")
                if results[func_name]:
                    logging.info(f"✅ Resultado de la validación {func_name} exitosa: {results[func_name]}")
                else:
                    logging.error(f"❌ Resultado de la validación {func_name} fallida: {results[func_name]}")
        return results



"""Módulo de encapsulamiento de módulos del subfolder."""

# Librerías Internas.
from migracion_gcp.lib.io.csv_file import CSVFile
from migracion_gcp.lib.io.json_file import JSONFile
from migracion_gcp.lib.io.pickle import PickleFile
from migracion_gcp.lib.io.registry import LocalFileRegistry


__all__ = ["CSVFile", "JSONFile", "PickleFile", "LocalFileRegistry"]


"""Módulo que define la interfaz para manejar archivos locales."""

# Librerías Externas.
from typing import Any, Optional
from abc import ABC, abstractmethod

import os
import shutil
import logging


class LocalFileBase(ABC):
    """Clase para manejar contratos de archivos locales."""

    def __init__(self, temp_dir: Optional[str] = "temp_dir") -> None:
        """Método para inicializar un archivo local.
        
        Args:
        ----------
        file_paths: List[str].
            Ruta del archivo que se desea manejar."""
        
        self.temp_dir = temp_dir

    def __enter__(self) -> "LocalFileBase":
        """Método para entrar en el contexto de un archivo local."""

        logging.info("Entrando en el contexto de un archivo local para hacerlo temporal.")

        os.makedirs(self.temp_dir, exist_ok = True)

        logging.info(f"Directorio temporal creado: {self.temp_dir}")

        return self
    
    def __exit__(self, exc_type: Any, exc_value: Any, traceback: Any) -> None:
        """Método para salir del contexto de un archivo local."""

        shutil.rmtree(self.temp_dir)

        logging.info("Saliendo del contexto de un archivo local temporal.")
    
    @abstractmethod
    def write_file(self, file_path: str, content: Any) -> None:
        """Método para escribir un archivo.
        
        Args:
        ----------
        file_path: str.
            Ruta del archivo que se desea escribir.

        content: Any.
            Contenido del archivo."""

        raise NotImplementedError("Si tu clase implementa el contrato LocalFileBase, debes implementar el método 'write_file'.")



"""Módulo que define un builder para la validación de modelos de regresión."""

# Librerías Externas.
from typing import Callable

import numpy as np
import pandas as pd

from scipy.stats import ks_2samp
from sklearn.base import BaseEstimator

# Librerías Internas.
from genesis_explorer.validations.builder import Builder


class ValidateRegressionBuilder(Builder):
    """Clase que implementa un builder para la validación de
    modelos de regresión."""
    
    def add_metrics_validation(self) -> "ValidateRegressionBuilder":
        """Método que define el contrato para la validación de métricas.
        
        Returns:
        ----------
        ValidateRegressionBuilder.
            Builder para la validación de métricas."""

        def validate_metrics(modelo: BaseEstimator,
                             metric_fn: Callable[[np.ndarray, np.ndarray], float],
                             X_train: pd.DataFrame,
                             X_test: pd.DataFrame,
                             y_train: pd.Series,
                             y_test: pd.Series, **kwargs) -> bool:
            """Función que se encarga de validar las métricas del modelo.

            Args:
            ----------
            modelo: BaseEstimator.
                Modelo de regresión.

            metric_fn: Callable[[np.ndarray, np.ndarray], float].
                Función que define la métrica a validar.
            
            X_train: pd.DataFrame.
            
            X_test: pd.DataFrame.
                Conjunto de datos de prueba.
            
            y_train: pd.Series.
                Conjunto de datos de entrenamiento.
            
            y_test: pd.Series.
                Conjunto de datos de prueba.
            
            Returns:
            ----------
            bool.
                True si las métricas del modelo son válidas, False en caso contrario."""
            
            try:
                train_scores = modelo.predict(X_train)
                test_scores = modelo.predict(X_test)

                train_metric = metric_fn(train_scores, y_train)
                test_metric = metric_fn(test_scores, y_test)
            
            except Exception as e:
                raise Exception(f"Error al validar las métricas del modelo: {e}")
                    
            val1 = train_metric < test_metric
            val2 = abs(train_metric - test_metric) < self._metrics_threshold * test_metric
            return val1 and val2
        
        self._validator.add_validation(validate_metrics)
        return self
    
    def add_scores_validation(self) -> "ValidateRegressionBuilder":
        """Método que define el contrato para la validación de scores.
        
        Returns:
        ----------
        ValidateRegressionBuilder.
            Builder para la validación de scores."""
        
        def validate_scores(y_train: pd.Series,
                            y_test: pd.Series, **kwargs) -> bool:
            """Función que se encarga de validar los scores del modelo.
            
            Args:
            ----------
            y_train: pd.Series.
                Conjunto de datos de entrenamiento.
            
            y_test: pd.Series.
                Conjunto de datos de prueba.
            
            Returns:
            ----------
            bool.
                True si los scores del modelo son válidos, False en caso contrario."""
            
            try:
                _, ks_pvalue = ks_2samp(y_train.values, y_test.values)
            except Exception as e:
                raise Exception(f"Error al validar los scores del modelo: {e}")
            
            return float(ks_pvalue) < self._pvalue_threshold
                        
        self._validator.add_validation(validate_scores)
        return self



"""Módulo que define el cliente para comunicación con Pub/Sub."""

# Librerías Externas.
from typing import Callable, Optional, List, Dict, Any

import logging

from google.cloud import pubsub_v1
from google.api_core.exceptions import AlreadyExists


# Librerías Internas.
from genesis_explorer.clients.base.base_messaging import BaseMessaging


logging.basicConfig(level = logging.INFO,
                    format = "%(asctime)s - %(levelname)s - %(message)s")


class PubSubService(BaseMessaging):
    """Clase que implementa el contrato de servicios de mensajería."""

    def __init__(self, project_id: str) -> None:
        """Método de instanciación de clase.
        
        Args:
        ----------
        project_id: str
            Identificador del proyecto de Google Cloud Platform."""
        
        self.project_id = project_id
        self.publisher = pubsub_v1.PublisherClient()
        self.subscriber = pubsub_v1.SubscriberClient()

    def create_topic(self, topic_name: str) -> None:
        """Método para crear un tópico de Pub/Sub.
        
        Args:
        ----------
        topic_name: str
            Nombre del tópico de Pub/Sub."""
        
        logging.info(f"Creando el tópico: {topic_name}.")
        topic_path = self.publisher.topic_path(project = self.project_id, topic = topic_name)

        try:
            self.publisher.create_topic(request = {"name": topic_path})
            logging.info(f"Tópico creado correctamente: {topic_path}.")
        except AlreadyExists:
            logging.info(f"El tópico {topic_name} ya existe.")
        except Exception as e:
            raise Exception(f"Error al crear el tópico de Pub/Sub: {e}")
        
    def subscribe(self, sub_name: str, topic_name: str) -> None:
        """Método para suscribirse a un tópico de Pub/Sub.
        
        Args:
        ----------
        sub_name: str
            Nombre de la suscripción de Pub/Sub.

        topic_name: str
            Nombre del tópico de Pub/Sub."""
        
        logging.info(f"Suscribiéndose al tópico: {topic_name}.")
        topic_path = self.publisher.topic_path(project = self.project_id, topic = topic_name)
        subscription_path = self.subscriber.subscription_path(project = self.project_id, subscription = sub_name)

        try:
            self.subscriber.create_subscription(request = {"name": subscription_path, "topic": topic_path})
            logging.info(f"Suscripción creada correctamente: {subscription_path}.")
        except AlreadyExists:
            logging.info(f"La suscripción {sub_name} ya existe.")
        except Exception as e:
            raise Exception(f"Error al crear la suscripción de Pub/Sub: {e}")
            
    def publish(self, topic_name: str, message: str) -> None:
        """Método para publicar un mensaje en un tópico de Pub/Sub.
        
        Args:
        ----------
        topic_name: str
            Nombre del tópico de Pub/Sub.

        message: str
            Mensaje a publicar."""
        
        logging.info(f"Publicando el mensaje en el tópico: {topic_name}.")
        topic_path = self.publisher.topic_path(project = self.project_id, topic = topic_name)

        try:
            encoded_message = str(message).encode("utf-8")
            future = self.publisher.publish(topic = topic_path, data = encoded_message)

            future.result()
            logging.info(f"Mensaje publicado correctamente en {topic_path}.")
        except Exception as e:
            logging.error(f"Error al publicar el mensaje en {topic_path}: {e}")

    def batch_pull(self, sub_name: str, max_messages: Optional[int] = 10) -> List[Dict[str, Any]]:
        """Método para obtener mensajes de un tópico de Pub/Sub en lotes.
        
        Args:
        ----------
        sub_name: str
            Nombre de la suscripción de Pub/Sub.

        max_messages: Optional[int]
            Número máximo de mensajes a obtener.

        Returns:
        ----------
        List[Dict[str, Any]].
            Lista de mensajes obtenidos."""
        
        logging.info(f"Obteniendo mensajes de la suscripción: {sub_name}.")

        subscription_path = self.subscriber.subscription_path(project = self.project_id,
                                                              subscription = sub_name)

        response = self.subscriber.pull(request = {"subscription": subscription_path,
                                                   "max_messages": max_messages})
        
        messages = []
        for msg in response.received_messages:
            messages.append({"ack_id": msg.ack_id,
                             "data": msg.message.data.decode("utf-8"),
                             "publish_time": msg.message.publish_time})
            
            logging.info(f"Marcando como leído el mensaje: {msg.ack_id}")
            self.subscriber.acknowledge(request = {"subscription": subscription_path,
                                                   "ack_ids": [msg.ack_id]})
            logging.info(f"Mensaje marcado como leído.")

        return messages
        
    def stream_pull(self, sub_name: str, callback: Callable) -> pubsub_v1.subscriber.futures.StreamingPullFuture:
        """Método para suscribirse a un tópico de Pub/Sub en streaming.
        
        Args:
        ----------
        sub_name: str
            Nombre de la suscripción de Pub/Sub.

        callback: Callable
            Función de callback para procesar los mensajes.

        Returns:
        ----------
        pubsub_v1.subscriber.futures.StreamingPullFuture.
            Cliente de suscripción de Pub/Sub."""

        subscription_path = self.subscriber.subscription_path(project = self.project_id, subscription = sub_name)
        streaming_pull_future = self.subscriber.subscribe(subscription = subscription_path, callback = callback)
        return streaming_pull_future

    def __str__(self) -> str:
        """Método para obtener la representación en cadena de caracteres de la clase.
        
        Returns:
        ----------
        str: Representación en cadena de caracteres de la clase."""

        self.publisher.list_topics(name = "mlops-credits-vertex-poc",)

        return f"PubSubService(project_id = {self.project_id})"



"""Módulo que define un cliente para la comunicación con Cloud Functions."""

# Librerías Externas.
from typing import Tuple, Optional

from google.cloud import functions_v2

# Librerías Internas.
from genesis_explorer.clients.base.base_functions import BaseFunction


class CloudFunction(BaseFunction):
    """Clase que define un cliente para la comunicación con Cloud Functions."""

    def __init__(self, project_id: str,
                 location: Optional[str] = "us-east1") -> None:
        """Método de instanciación de clase.
        
        Args:
        ----------
        project_id: str.
            ID del proyecto de GCP sobre el cual se creará el cliente.

        location: str.
            Región de GCP donde se creará la función."""

        self.project_id = project_id
        self.parent = f"projects/{project_id}/locations/{location}"

        self.functions_client = functions_v2.FunctionServiceClient()

    def create_function(self,
                        fn_name: str,
                        trigger_type: str,
                        entry_point: str,
                        source_archive_url: str,
                        runtime: Optional[str] = "python312",
                        trigger_topic_name: Optional[str] = None,
                        trigger_bucket_name: Optional[str] = None,
                        environment_variables: Optional[dict] = None) -> None:
        """Método que, dada la ruta a su archivo zip (que contiene el script de la función
        y sus requirements) en GCS, crea una Cloud Function en GCP.
        
        Args:
        ----------
        fn_name: str.
            Nombre de la función.
            
        trigger_type: str.
            Tipo de trigger de la función.
            
        entry_point: str.
            Nombre de la función en el script que la define.
            
        source_archive_url: str.
            URL del archivo zip que contiene el script de la función y sus requirements.
            
        bucket_name: str.
            Nombre del bucket de GCS donde se encuentra el archivo zip.
            
        runtime: str.
            Runtime de la función.
            
        topic_name: str.
            Nombre del topic de PubSub. Se usa si su trigger es de tipo PubSub.
            
        bucket_name: str.
            Nombre del bucket de GCS donde se encuentra el archivo zip. Se usa si su trigger es de tipo Storage.
            
        environment_variables: dict.
            Variables de entorno de la función."""

        function_name = f"{self.parent}/functions/{fn_name}"

        bucket_name, object_name = CloudFunction.__split_source_archive_url(source_archive_url)

        build_config = functions_v2.BuildConfig(runtime = runtime,
                                                entry_point = entry_point,
                                                environment_variables = environment_variables or {},
                                                source = functions_v2.Source(storage_source = functions_v2.StorageSource(bucket = bucket_name,
                                                                                                                         object_ = object_name)))

        if trigger_type == "http":
            event_trigger = None

        elif trigger_type == "pubsub":
            if trigger_topic_name is None:
                raise ValueError("El nombre del topic es obligatorio.")
            
            event_trigger = functions_v2.EventTrigger(event_type = "google.cloud.pubsub.topic.v1.messagePublished",
                                                      pubsub_topic = f"projects/{self.project_id}/topics/{trigger_topic_name}")
            
        elif trigger_type == "storage":
            if trigger_bucket_name is None:
                raise ValueError("El nombre del bucket es obligatorio.")
            
            event_trigger = functions_v2.EventTrigger(event_type = "google.storage.object.finalize",
                                                      event_filters = [functions_v2.EventFilter(attribute = "bucket",
                                                                                                value = trigger_bucket_name)])

        else:
            raise ValueError(f"El tipo de trigger {trigger_type} no es válido.")

        function = functions_v2.Function(name = function_name,
                                         description = "Trigger de entrenamiento del modelo.",
                                         build_config = build_config,
                                         event_trigger = event_trigger)

        operation = functions_v2.CreateFunctionRequest(parent = self.parent,
                                                       function = function,
                                                       function_id = fn_name)

        operation = self.functions_client.create_function(request = operation)
        operation.result()
    
    @staticmethod
    def __split_source_archive_url(source_archive_url: str) -> Tuple[str, str]:
        """Método estático y privado que, dado el URL de un archivo zip en GCS,
        devuelve el nombre del bucket y el nombre del objeto.
        
        Args:
        ----------
        source_archive_url: str.
            URL del archivo zip que contiene el script de la función y sus requirements.

        Returns:
        ----------
        bucket_name: str.
            Nombre del bucket de GCS donde se encuentra el archivo zip.
            
        object_name: str.
            Nombre del objeto en el bucket de GCS donde se encuentra el archivo zip."""

        bucket_name, object_name = source_archive_url.replace("gs://", "").split("/", 1)
        return bucket_name, object_name



"""Módulo que define la clase para manejar archivos CSV locales."""

# Librerías Externas.
import logging
import pandas as pd

# Librerías Internas.
from genesis_explorer.io.base import LocalFileBase


logging.basicConfig(level = logging.INFO, 
                    format = "%(asctime)s - %(levelname)s - %(message)s")


class CSVFile(LocalFileBase):
    """Clase para manejar archivos CSV locales."""
    
    def write_file(self, file_path: str, content: pd.DataFrame, **kwargs) -> None:
        """Método para escribir un archivo CSV."""

        logging.info(f"Escribiendo el archivo {file_path} en el directorio temporal...")

        try:
            content.to_csv(f"{self.temp_dir}/{file_path}", **kwargs)

            logging.info(f"Archivo {file_path} escrito correctamente.")

        except Exception as e:
            logging.error(f"Error al escribir el archivo {file_path}: {e}")



"""Módulo que define la clase para manejar archivos pickle locales."""


# Librerías Externas.
from typing import Any

import pickle
import logging

# Librerías Internas.
from genesis_explorer.io.base import LocalFileBase
from genesis_explorer.io.registry import LocalFileRegistry


@LocalFileRegistry.register_file_handler("pickle")
class PickleFile(LocalFileBase):
    """Clase para manejar archivos pickle locales."""

    def write_file(self,
                   file_path: str,
                   content: Any,
                   **kwargs) -> None:
        """Método para escribir un archivo pickle.
        
        Args:
        ----------
        file_path: str.
            Ruta del archivo que se desea escribir.
        
        content: Any.
            Contenido del archivo."""
        
        logging.info(f"Escribiendo el archivo {file_path} en el directorio temporal...")

        try:
            with open(f"{self.temp_dir}/{file_path}", "wb") as file:
                pickle.dump(content, file, **kwargs)

            logging.info(f"Archivo {file_path} escrito correctamente.")

        except Exception as e:
            logging.error(f"Error al escribir el archivo {file_path}: {e}")
